# Tool names
BLOBS = 'blobs'
CALCULATED_PROPERTIES = 'calculated_properties'
COLLECTIONS = 'collections'
CONNECTION = 'connection'
DBSESSION = 'dbsession'
STORAGE = 'storage'
ROOT = 'root'
TYPES = 'types'
UPGRADER = 'upgrader'

# Constants
PHASE1_5_CONFIG = -15
PHASE2_5_CONFIG = -5


# Events
class Created(object):
    def __init__(self, object, request):
        self.object = object
        self.request = request


class BeforeModified(object):
    def __init__(self, object, request):
        self.object = object
        self.request = request


class AfterModified(object):
    def __init__(self, object, request):
        self.object = object
        self.request = request


class AfterUpgrade(object):
    def __init__(self, object):
        self.object = object



import pytest


targets = [
    {'name': 'one', 'uuid': '775795d3-4410-4114-836b-8eeecf1d0c2f'},
    {'name': 'two', 'uuid': 'd6784f5e-48a1-4b40-9b11-c8aefb6e1377'},
]

item = {
    'required': 'required value',
}

simple1 = {
    'required': 'required value',
    'simple1': 'supplied simple1',
}

simple2 = {
    'required': 'required value',
    'simple2': 'supplied simple2',
}

item_with_uuid = [
    {
        'uuid': '0f13ff76-c559-4e70-9497-a6130841df9f',
        'required': 'required value 1',
        'field_no_default': 'test'

    },
    {
        'uuid': '6c3e444b-f290-43c4-bfb9-d20135377770',
        'required': 'required value 2',
    },
]

item_with_link = [
    {
        'required': 'required value 1',
        'protected_link': '775795d3-4410-4114-836b-8eeecf1d0c2f',
    },
    {
        'required': 'required value 2',
        'protected_link': 'd6784f5e-48a1-4b40-9b11-c8aefb6e1377',
    },
]


COLLECTION_URL = '/testing-post-put-patch-sno/'


@pytest.fixture
def link_targets(testapp):
    url = '/testing-link-targets-sno/'
    for item in targets:
        testapp.post_json(url, item, status=201)


@pytest.fixture
def content(testapp, external_tx):
    res = testapp.post_json(COLLECTION_URL, item_with_uuid[0], status=201)
    return {'@id': res.location}


@pytest.fixture
def content_with_child(testapp):
    parent_res = testapp.post_json('/testing-link-targets-sno/', {}, status=201)
    parent_id = parent_res.json['@graph'][0]['@id']
    child_res = testapp.post_json('/testing-link-sources-sno/', {'target': parent_id})
    child_id = child_res.json['@graph'][0]['@id']
    return {'@id': parent_id, 'child': child_id}


def test_admin_post(testapp, external_tx):
    testapp.post_json(COLLECTION_URL, item, status=201)
    testapp.post_json(COLLECTION_URL, item_with_uuid[0], status=201)


def test_admin_put_uuid(content, testapp):
    url = content['@id']
    # so long as the same uuid is supplied, PUTing the uuid is fine
    testapp.put_json(url, item_with_uuid[0], status=200)
    # but the uuid may not be changed on PUT;
    testapp.put_json(url, item_with_uuid[1], status=422)


def test_defaults_on_put(content, testapp):
    url = content['@id']
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'
    assert res.json['simple2'] == 'simple2 default'

    res = testapp.put_json(url, simple1, status=200)
    assert res.json['@graph'][0]['simple1'] == 'supplied simple1'
    assert res.json['@graph'][0]['simple2'] == 'simple2 default'

    res = testapp.put_json(url, simple2, status=200)
    assert res.json['@graph'][0]['simple1'] == 'simple1 default'
    assert res.json['@graph'][0]['simple2'] == 'supplied simple2'


def test_patch(content, testapp):
    url = content['@id']
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'
    assert res.json['simple2'] == 'simple2 default'

    res = testapp.patch_json(url, {}, status=200)
    assert res.json['@graph'][0]['simple1'] == 'simple1 default'
    assert res.json['@graph'][0]['simple2'] == 'simple2 default'

    res = testapp.patch_json(url, {'simple1': 'supplied simple1'}, status=200)
    assert res.json['@graph'][0]['simple1'] == 'supplied simple1'
    assert res.json['@graph'][0]['simple2'] == 'simple2 default'

    res = testapp.patch_json(url, {'simple2': 'supplied simple2'}, status=200)
    assert res.json['@graph'][0]['simple1'] == 'supplied simple1'
    assert res.json['@graph'][0]['simple2'] == 'supplied simple2'


def test_patch_delete_fields(content, testapp):
    url = content['@id']
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'
    assert res.json['simple2'] == 'simple2 default'
    assert res.json['field_no_default'] == 'test'

    res = testapp.patch_json(url, {'simple1': 'this is a test'}, status=200)
    assert res.json['@graph'][0]['simple1'] == 'this is a test'

    # delete fields with defaults resets to default, while deleting non default field
    # completely removes them
    res = testapp.patch_json(url + "?delete_fields=simple1,field_no_default", {}, status=200)
    assert 'field_no_default' not in res.json['@graph'][0].keys()
    assert res.json['@graph'][0]['simple1'] == 'simple1 default'


def test_patch_delete_fields_enum(content, testapp):
    """
    enum with no default set, should delete completely
    """
    url = content['@id']
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'
    assert res.json['simple2'] == 'simple2 default'

    res = testapp.patch_json(url, {'enum_no_default': '2', 'simple1': 'this is a test'}, status=200)
    assert res.json['@graph'][0]['enum_no_default'] == '2'
    assert res.json['@graph'][0]['simple1'] == 'this is a test'

    # delete enums without a default removes it completely
    res = testapp.patch_json(url + "?delete_fields=simple1,enum_no_default", {}, status=200)
    assert 'enum_no_default' not in res.json['@graph'][0].keys()
    assert res.json['@graph'][0]['simple1'] == 'simple1 default'


def test_patch_delete_fields_non_string(content, testapp):
    url = content['@id']
    res = testapp.get(url)

    # delete fields with defaults resets to default, while deleting non default field
    # completely removes them
    res = testapp.patch_json(url + "?delete_fields=schema_version", {}, status=200)
    assert res.json['@graph'][0]['schema_version'] == '1'


def test_patch_delete_fields_fails_with_no_validation(content, testapp):
    url = content['@id']
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'
    assert res.json['simple2'] == 'simple2 default'
    assert res.json['field_no_default'] == 'test'

    # using delete_fields with validate=False will now raise a validation err
    res = testapp.patch_json(url + "?delete_fields=simple1,field_no_default&validate=false", {}, status=422)
    assert res.json['description'] == "Failed validation"
    assert 'Cannot delete fields' in res.json['errors'][0]['description']


def test_patch_delete_fields_bad_param(content, testapp):
    """
    delete_fields should not fail with a bad fieldname, but simply ignore
    """
    url = content['@id']
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'
    assert res.json['simple2'] == 'simple2 default'
    assert res.json['field_no_default'] == 'test'
    res = testapp.patch_json(url + "?delete_fields=simple1,bad_fieldname", {}, status=200)
    # default value
    assert res.json['@graph'][0]['simple1'] == 'simple1 default'
    assert 'bad_fieldname' not in res.json['@graph'][0]


def test_patch_delete_fields_import_items_admin(link_targets, testapp):
    res = testapp.post_json(COLLECTION_URL, item_with_link[0], status=201)
    url = res.location
    assert res.json['@graph'][0]['protected_link']
    res = testapp.patch_json(url + "?delete_fields=protected_link", {}, status=200)


def test_patch_delete_fields_required(content, testapp):
    url = content['@id']
    res = testapp.get(url)

    # with validate=false, then defaults are not populated so default fields are also deleted
    res = testapp.patch_json(url + "?delete_fields=required", {}, status=422)
    assert res.json['description'] == "Failed validation"
    assert res.json['errors'][0]['name'] == "Schema: "
    assert res.json['errors'][0]['description'] == "'required' is a required property"


def test_patch_new_schema_version(content, root, testapp, monkeypatch):
    collection = root['testing_post_put_patch_sno']
    properties = collection.type_info.schema['properties']

    url = content['@id']
    res = testapp.get(url)
    assert res.json['schema_version'] == '1'

    monkeypatch.setitem(properties['schema_version'], 'default', '2')
    monkeypatch.setattr(collection.type_info, 'schema_version', '2')
    monkeypatch.setitem(properties, 'new_property', {'default': 'new'})
    res = testapp.patch_json(url, {}, status=200)
    assert res.json['@graph'][0]['schema_version'] == '2'
    assert res.json['@graph'][0]['new_property'] == 'new'


def test_admin_put_protected_link(link_targets, testapp):
    res = testapp.post_json(COLLECTION_URL, item_with_link[0], status=201)
    url = res.location

    testapp.put_json(url, item_with_link[0], status=200)
    testapp.put_json(url, item_with_link[1], status=200)


def test_put_object_not_touching_children(content_with_child, testapp):
    url = content_with_child['@id']
    res = testapp.put_json(url, {}, status=200)
    assert content_with_child['child'] in res.json['@graph'][0]['reverse']


def test_put_object_editing_child(content_with_child, testapp):
    edit = {
        'reverse': [{
            '@id': content_with_child['child'],
            'status': 'released',
        }]
    }
    # this is no longer allowed
    res = testapp.put_json(content_with_child['@id'], edit, status=422)
    assert res.json['description'] == 'Failed validation'
    assert len(res.json['errors']) == 1
    res_error = res.json['errors'][0]
    assert res_error['name'] == 'Schema: reverse'
    assert res_error['description'] == 'submission of calculatedProperty disallowed'


def test_name_key_validation(link_targets, testapp):
    # name_key
    target_data = {'name': 'one#name'}
    res = testapp.post_json('/testing-link-targets-sno/', target_data, status=422)
    assert res.json['description'] == 'Failed validation'
    res_error = res.json['errors'][0]
    assert res_error['name'] == "Item: path characters"
    assert "Forbidden character(s) {'#'}" in res_error['description']

    # unique_key
    source_data = {'name': 'two@*name', 'target': targets[0]['uuid']}
    res = testapp.post_json('/testing-link-sources-sno/', source_data, status=422)
    assert res.json['description'] == 'Failed validation'
    res_error = res.json['errors'][0]
    assert res_error['name'] == 'Item: path characters'
    assert "Forbidden character(s) {'*'}" in res_error['description']


def test_retry(testapp):
    res = testapp.post_json(COLLECTION_URL, {'required': ''})
    url = res.location
    res = testapp.get(url + '/@@testing-retry?datastore=database')
    assert res.json['attempt'] == 2
    assert not res.json['detached']


def test_check_only(content, testapp):
    url = content['@id']
    # PATCH
    res = testapp.patch_json(url + '?check_only=True',
                             {'simple1': 'supplied simple1'}, status=200)
    assert res.json['status'] == 'success'
    assert res.json['@type'] == ['result']
    # item did not change
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'

    # PUT
    altered_item = item_with_uuid[0].copy()
    altered_item['simple1'] = 'supplied simple1'
    res = testapp.put_json(url+ '?check_only=True', item_with_uuid[0], status=200)
    assert res.json['status'] == 'success'
    assert res.json['@type'] == ['result']
    # item did not change
    res = testapp.get(url)
    assert res.json['simple1'] == 'simple1 default'

    # POST
    res = testapp.post_json(COLLECTION_URL + '?check_only=True', {'required': ''})
    assert res.json['status'] == 'success'
    assert res.json['@type'] == ['result']

    # check_only raises validation errors
    res = testapp.patch_json(url + "?check_only=True&delete_fields=required", {}, status=422)
    assert len(res.json['errors']) == 2
    val_error = res.json['errors'][0]
    assert val_error['name'] == 'Schema: '
    assert val_error['description'] == "'required' is a required property"
    del_error = res.json['errors'][1]
    assert del_error['name'] == 'delete_fields'
    assert del_error['description'] == 'Error deleting fields'


def test_max_sid_view(content, testapp):
    res = testapp.get('/max-sid', status=200)
    assert res.json['status'] == 'success'
    assert 'max_sid' in res.json
    starting_sid = res.json['max_sid']
    # increment sid and make sure it is updated
    url = content['@id']
    testapp.patch_json(url, {}, status=200)

    res = testapp.get('/max-sid', status=200)
    assert res.json['status'] == 'success'
    assert res.json['max_sid'] > starting_sid


def test_create_es_item_without_es(content, testapp):
    """
    Items with `properties_datastore='elasticsearch'` should fail without ES set up
    """
    target_data = {'name': 'es_target_test'}
    res = testapp.post_json('/testing-link-targets-elastic-search/', target_data, status=500)
    assert res.json['detail'] == 'Forced datastore elasticsearch is not configured'



import structlog
from elasticsearch.helpers import scan
from ..interfaces import COLLECTIONS, TYPES
from .interfaces import ELASTIC_SEARCH
from ..util import DEFAULT_EMBEDS, crawl_schema


log = structlog.getLogger(__name__)


def get_namespaced_index(config, index):
    """ Grabs indexer.namespace from settings and namespace the given index """
    try:
        settings = config.registry.settings
    except:  # accept either config or registry as first arg
        settings = config.settings
    namespace = settings.get('indexer.namespace') or ''
    return namespace + index


def namespace_index_from_health(health, index):
    """ Namespaces the given index based on health page data """
    if 'error' in health:
        raise RuntimeError('Mirror health unresolved: %s' % health)
    return health.get('namespace', '') + index


def to_camel_case(snake_string):
    return snake_string.title().replace("_", "")


def find_uuids_for_indexing(registry, updated, find_index=None):
    """
    Run a search to find uuids of objects with that contain the given set of
    updated uuids in their linked_uuids.
    Uses elasticsearch.helpers.scan to iterate through ES results.
    Returns a set containing original uuids and the found uuids (INCLUDING
    uuids that were passed into this function)

    Args:
        registry: the current Registry
        updated (set): uuids to use as basis for finding associated items
        find_index (str): index to search in. Default to '_all' (all indices)

    Return:
        set: of uuids, including associated uuids found AND `updated` uuids
    """
    es = registry[ELASTIC_SEARCH]
    scan_query = {
        'query': {
            'bool': {
                'filter': {
                    'bool': {
                        'should': [
                            {
                                'terms': {
                                    'linked_uuids_embedded.uuid': list(updated)
                                }
                            }
                        ]
                    }
                }
            }
        },
        '_source': False
    }
    if not find_index:
        find_index = get_namespaced_index(registry, '*')
    results = scan(es, index=find_index, query=scan_query)
    invalidated_with_type = {(res['_id'], to_camel_case(res['_type'])) for res in results}
    invalidated = {uuid for uuid, type in invalidated_with_type}

    return updated | invalidated, invalidated_with_type


def get_uuids_for_types(registry, types=[]):
    """
    WARNING! This makes lots of DB requests and should be used carefully.

    Generator function to return uuids for all the given types. If no
    types provided, uses all types (get all uuids). Because of inheritance
    between item classes, do not iterate over all subtypes (as is done with
    `for uuid in collection`; instead, leverage `collection.iter_no_subtypes`)

    Args:
        registry: the current Registry
        types (list): string item types to specifcally use to find collections.
            Default is empty list, which means all collections are used

    Yields:
        str: uuid of item in collections
    """
    if not isinstance(types, list) or not all(isinstance(t, str) for t in types):  # type check for safety
        raise TypeError('Expected type=list (of strings) for argument "types"')
    collections = registry[COLLECTIONS]
    # might as well sort collections alphatbetically, as this was done earlier
    for coll_name in sorted(collections.by_item_type):
        if types and coll_name not in types:
            continue
        for uuid in collections.by_item_type[coll_name].iter_no_subtypes():
            yield str(uuid)


def extract_type_properties(registry, invalidated_item_type):
    """ Helper function, useful for mocking. """
    return registry['types'][invalidated_item_type].schema['properties']


def extract_type_embedded_list(registry, invalidated_item_type):
    """ Helper function, useful for mocking """
    return registry['types'][invalidated_item_type].embedded_list


def extract_base_types(registry, item_type):
    """ Helper function, useful for mocking """
    return registry[TYPES][item_type].base_types


def determine_parent_types(registry, item_type):
    """ Determines the parent types of the given item_type """
    base_types = []
    try:
        base_types = extract_base_types(registry, item_type)
        if base_types == ['Item']:  # trivially true of all items
            return []
    except KeyError:  # indicative of an error if not testing
        log.info('Tried to determine parent type of invalid type: %s' % item_type)
    return [b for b in base_types if b != 'Item']


def build_diff_metadata(registry, diff):
    """ Helper function for below that builds metadata from diff needed to filter
        invalidation scope.

    :param registry: application registry, used to retrieve type information
    :param diff: a diff of the change (from SQS), see build_diff_from_request
    :returns: 3-tuple skip bool (to invalidate everything),
              dictionaries of diff intermediary and child -> parent type mappings
    """
    # build representation of diffs
    # item type -> modified fields mapping
    diffs, child_to_parent_type = {}, {}
    skip = False  # if a modified field is a default embed, EVERYTHING has to be invalidated
    for _d in diff:
        modified_item_type, modified_field = _d.split('.', 1)
        if ('.' + modified_field) in DEFAULT_EMBEDS + ['.status']:  # XXX: 'status' is an unnamed default embed?
            skip = True
            break
        if modified_item_type not in diffs:
            diffs[modified_item_type] = [modified_field]
        else:
            diffs[modified_item_type].append(modified_field)

        modified_item_parent_types = determine_parent_types(registry, modified_item_type)
        if modified_item_parent_types:
            child_to_parent_type[modified_item_type] = modified_item_parent_types

    return skip, diffs, child_to_parent_type


def filter_invalidation_scope(registry, diff, invalidated_with_type, secondary_uuids):
    """ Function that given a diff in the following format:
            ItemType.base_field.terminal_field --> {ItemType: base_field.terminal_field} intermediary
        And a list of invalidated uuids with their type information as a 2-tuple:
            [(<uuid>, <item_type>), ...]
        Removes uuids of item types that were not invalidated from the set of secondary_uuids.

    :param registry: application registry, used to retrieve type information
    :param diff: a diff of the change (from SQS), see build_diff_from_request
    :param invalidated_with_type: list of 2-tuple (uuid, item_type)
    :param secondary_uuids: primary set of uuids to be invalidated
    """
    skip, diffs, child_to_parent_type = build_diff_metadata(registry, diff)

    # go through all invalidated uuids, looking at the embedded list of the item type
    item_type_is_invalidated = {}
    for invalidated_uuid, invalidated_item_type in invalidated_with_type:
        if skip is True:  # if we detected a change to a default embed, invalidate everything
            break

        # remove this uuid if its item type has been seen before and found to
        # not be invalidated
        if invalidated_item_type in item_type_is_invalidated:
            if item_type_is_invalidated[invalidated_item_type] is False:
                secondary_uuids.discard(invalidated_uuid)
            continue  # nothing else to do here

        # if we get here, we are looking at an invalidated_item_type that exists in the
        # diff and we need to inspect the embedded list to see if the diff fields are
        # embedded
        properties = extract_type_properties(registry, invalidated_item_type)
        embedded_list = extract_type_embedded_list(registry, invalidated_item_type)
        for embed in embedded_list:

            # check the field up to the embed as this is the path to the linkTo
            # we must determine it's type and determine if the given diff could've
            # resulted in an invalidation
            split_embed = embed.split('.')
            base_field, terminal_field = '.'.join(split_embed[0:-1]), split_embed[-1]
            base_field_schema = crawl_schema(registry['types'], base_field, properties)
            base_field_item_type = base_field_schema.get('linkTo', None)

            # recursive helper function that will drill down as much as necessary
            def locate_link_to(schema_cursor):
                if 'items' in schema_cursor:  # array
                    if 'properties' in schema_cursor['items']:
                        for field_name, details in schema_cursor['items']['properties'].items():
                            if base_field.endswith(field_name):
                                if 'linkTo' in details:
                                    return details['linkTo']
                                else:
                                    return locate_link_to(details)
                    else:
                        return schema_cursor['items']['linkTo']
                elif 'properties' in schema_cursor:  # object
                    for field_name, details in schema_cursor['properties'].items():
                        if base_field.endswith(field_name):
                            if 'linkTo' in details:
                                return details['linkTo']
                            else:
                                return locate_link_to(details)
                else:
                    log.error(schema_cursor)
                    raise Exception('Unexpected')

            # if we are not a top level linkTo, drill down
            if base_field_item_type is None:
                base_field_item_type = locate_link_to(base_field_schema)

            # Collect diffs from all possible item_types
            all_possible_diffs = diffs.get(base_field_item_type, [])
            parent_types = child_to_parent_type.get(base_field_item_type, None)
            if parent_types is not None:
                all_possible_diffs += [diffs.get(parent) for parent in parent_types]

            if not all_possible_diffs:  # no diffs match this embed
                continue

            # XXX VERY IMPORTANT: for this to work correctly, the fields used in calculated properties MUST
            # be embedded! In addition, if you embed * on a linkTo, modifications to that linkTo will ALWAYS
            # invalidate the item_type
            if (any(terminal_field == field for field in all_possible_diffs) or
                    terminal_field.endswith('*')):
                item_type_is_invalidated[invalidated_item_type] = True
                break

        # if we didnt break out of the above loop, we never found an embedded field that was
        # touched, so set this item type to False so all items of this type are NOT invalidated
        if invalidated_item_type not in item_type_is_invalidated:
            secondary_uuids.discard(invalidated_uuid)
            item_type_is_invalidated[invalidated_item_type] = False

    # XXX: Enable to get debugging information on invalidation scope
    # def _sort(tp):
    #     return tp[0]
    # log.error('Diff: %s Invalidated: %s Cleared: %s' % (diffs, sorted(list((k, v) for k, v in item_type_is_invalidated.items()
    #                                                                        if v is True), key=_sort),
    #                                                            sorted(list((k, v) for k, v in item_type_is_invalidated.items()
    #                                                                        if v is False), key=_sort)))



import copy
import datetime
import json
import time
from timeit import default_timer as timer

import structlog
from elasticsearch.exceptions import (
    ConflictError,
    ConnectionError,
    TransportError,
)
from pyramid.view import view_config
from urllib3.exceptions import ReadTimeoutError

from ..interfaces import (
    DBSESSION,
    STORAGE
)
from .indexer_utils import get_namespaced_index, find_uuids_for_indexing, filter_invalidation_scope
from .interfaces import (
    ELASTIC_SEARCH,
    INDEXER,
    INDEXER_QUEUE
)
from ..embed import MissingIndexItemException
from ..util import debug_log, dictionary_lookup


log = structlog.getLogger(__name__)


def includeme(config):
    config.add_route('index', '/index')
    config.scan(__name__)
    registry = config.registry
    registry[INDEXER] = Indexer(registry)


# really simple exception to know when the sid check fails
class SidException(Exception):
    pass


def check_sid(sid, max_sid):
    """
    Simple function to compare a given sid to given max_sid.
    Raise an Exception if malformed or lesser max_sid

    Args:
        sid (int): query sid
        max_sid (int): maximum sid to compare to

    Raises:
        ValueError: if sid or max_sid are not valid
        SidException: if sid in request is greater than max sid
    """
    try:
        sid = int(sid)
        max_sid = int(max_sid)
    except ValueError:
        raise ValueError('sid (%s) and max sid (%s) must be integers.'
                         % (sid, max_sid))
    if max_sid < sid:
        raise SidException('Query sid (%s) is greater than max sid (%s).'
                           % (sid, max_sid))


@view_config(route_name='index', request_method='POST', permission="index")
@debug_log
def index(context, request):
    # Setting request.datastore here only works because routed views are not traversed.
    request.datastore = 'database'
    record = request.json.get('record', False)  # if True, make a record in es
    dry_run = request.json.get('dry_run', False)  # if True, do not actually index
    es = request.registry[ELASTIC_SEARCH]
    indexer = request.registry[INDEXER]
    namespace_star = get_namespaced_index(request, '*')
    namespaced_index = get_namespaced_index(request, 'indexing')

    if not dry_run:
        index_start_time = datetime.datetime.now()
        index_start_str = index_start_time.isoformat()

        # create indexing record, with _id equal to starting timestamp
        indexing_record = {
            'uuid': index_start_str,
            'indexing_status': 'started',
        }

        # get info on what actually is being indexed
        indexing_content = {
            'type': 'sync' if request.json.get('uuids') else 'queue',
        }
        if indexing_content['type'] == 'sync':
            indexing_content['sync_uuids'] = len(request.json.get('uuids'))
        else:
            indexing_content['initial_queue_status'] = indexer.queue.number_of_messages()
        indexing_record['indexing_content'] = indexing_content
        indexing_record['indexing_started'] = index_start_str
        indexing_counter = [0]
        # actually index
        # try to ensure ES is reasonably up to date
        es.indices.refresh(index=namespace_star)

        # NOTE: the refresh interval is left as default because it doesn't seem
        # to help performance much.
        # However, disabling it is okay, since check_es_and_cache_linked_sids
        # uses a GET, which will call a refresh if needed
        # Enabling the line below adds ~1.5 second overhead before and after indexing
        # es.indices.put_settings(index='_all', body={'index' : {'refresh_interval': '-1'}})

        # do the indexing!
        indexing_record['errors'] = indexer.update_objects(request, indexing_counter)

        # get some final info for the record
        index_finish_time = datetime.datetime.now()
        indexing_record['indexing_finished'] = index_finish_time.isoformat()
        indexing_record['indexing_elapsed'] = str(index_finish_time - index_start_time)
        # update record with final queue snapshot
        if indexing_content['type'] == 'queue':
            indexing_content['finished_queue_status'] = indexer.queue.number_of_messages()
        indexing_record['indexing_count'] = indexing_counter[0]
        indexing_record['indexing_status'] = 'finished'

        # with the index listener running more frequently, we don't want to
        # store a ton of useless records. Only store queue records that have
        # errors or have non-zero indexing count
        if record and indexing_content['type'] == 'queue' and not indexing_record['errors']:
            record = indexing_record['indexing_count'] > 0

        if record:
            try:
                es.index(index=namespaced_index, doc_type='indexing', body=indexing_record, id=index_start_str)
                es.index(index=namespaced_index, doc_type='indexing', body=indexing_record, id='latest_indexing')
            except:
                indexing_record['indexing_status'] = 'errored'
                error_messages = copy.deepcopy(indexing_record['errors'])
                del indexing_record['errors']
                es.index(index=namespaced_index, doc_type='indexing', body=indexing_record, id=index_start_str)
                es.index(index=namespaced_index, doc_type='indexing', body=indexing_record, id='latest_indexing')
                for item in error_messages:
                    if 'error_message' in item:
                        log.error('Indexing error', **item)
                        item['error_message'] = "Error occured during indexing, check the logs"

        # this will make documents in all lucene buffers available to search
        es.indices.refresh(index=namespace_star)
        # resets the refresh_interval to the default value (must reset if disabled earlier)
        # es.indices.put_settings(index='_all', body={'index' : {'refresh_interval': '1s'}})
    return indexing_record


class Indexer(object):
    def __init__(self, registry):
        self.registry = registry
        self.es = registry[ELASTIC_SEARCH]
        self.queue = registry[INDEXER_QUEUE]

    def update_objects(self, request, counter):
        """
        Top level routing between `Indexer.update_objects_sync` (synchronous)
        and `Indexer.update_objects_queue` (asynchronous, usually used).
        Also sets isolation level for the DB connection
        """
        session = request.registry[DBSESSION]()
        connection = session.connection()
        connection.execute('SET TRANSACTION ISOLATION LEVEL REPEATABLE READ READ ONLY')

        # indexing is either run with sync uuids passed through the request
        # (which is synchronous) OR uuids from the queue
        sync_uuids = request.json.get('uuids', None)

        # actually index
        if sync_uuids:
            errors = self.update_objects_sync(request, sync_uuids, counter)
        else:
            errors, _ = self.update_objects_queue(request, counter)
        return errors

    def get_messages_from_queue(self):
        """
        Simple helper method. Attempt to get items from deferred queue first,
        and if none are found, check primary and then secondary queues. Both use
        long polling. Returns list of messages received and the string target
        the the queue came from.
        """
        messages = None
        target_queue = None
        for try_queue in self.queue.queue_targets:
            messages = self.queue.receive_messages(target_queue=try_queue)
            if messages:
                target_queue = try_queue
                break
        return messages, target_queue

    def find_and_queue_secondary_items(self, source_uuids, rev_linked_uuids,
                                       sid=None, telemetry_id=None, diff=None):
        """
        Find all associated uuids of the given set of non-strict uuids using ES
        and queue them in the secondary queue. Associated uuids include uuids
        that linkTo or are rev_linked to a given item.
        Add rev_linked_uuids linking to source items found from @@indexing-view
        after finding secondary uuids (they are "strict")
        """
        # find_uuids_for_indexing() will return items linking to and items
        # rev_linking to this item currently in ES (find old rev_links)
        associated_uuids, invalidated_with_type = find_uuids_for_indexing(self.registry, source_uuids)

        # remove already indexed primary uuids used to find them
        secondary_uuids = associated_uuids - source_uuids

        # we are updating from an edit and have a corresponding diff
        if diff is not None:
            filter_invalidation_scope(self.registry, diff, invalidated_with_type, secondary_uuids)

        # update this with rev_links found from @@indexing-view (includes new rev_links)
        # AFTER invalidation scope filtering, since invalidation scope does not account for rev-links
        secondary_uuids |= rev_linked_uuids

        # items queued through this function are ALWAYS strict in secondary queue
        return self.queue.add_uuids(self.registry, list(secondary_uuids), strict=True,
                                    target_queue='secondary', sid=sid,
                                    telemetry_id=telemetry_id)

    def update_objects_queue(self, request, counter):
        """
        Used for asynchronous indexing with the indexer queues. Some notes:
        - Keep track of `max_sid` of the transaction scope to defer out-of-scope
          indexing to a new run.
        - Iterate through queue messages, calling `update_object` on each
        - Handle deleting and recycling messages to the queues on errors/defers
        - Handles getting messages by priority with `get_messages_from_queue`
        """
        errors = []
        # hold uuids that will be used to find secondary uuids
        non_strict_uuids = set()
        # hold the reverse-linked uuids that need to be invalidated
        rev_linked_uuids = set()
        to_delete = []  # hold messages that will be deleted
        to_defer = []  # hold messages once we need to restart the worker
        # max_sid does not change of the lifetime of request
        max_sid = request.registry[STORAGE].write.get_max_sid()
        deferred = False  # if true, we need to restart the worker
        messages, target_queue = self.get_messages_from_queue()
        while len(messages) > 0:
            for idx, msg in enumerate(messages):
                msg_body = json.loads(msg['Body'])

                # handle case where worker needs to restart
                # recycle additional messages (no effect on dlq count) and
                # delete old messages
                if deferred:
                    to_defer.append(msg_body)
                    to_delete.append(msg)
                    if len(to_delete) == self.queue.delete_batch_size:
                        self.queue.delete_messages(to_delete, target_queue=target_queue)
                        to_delete = []
                    continue

                # This rather than msg_body['uuid'] to get better error reporting
                # in case uuid is missing from dictionary, which probably happens
                # because the msg_body is some other kind of object entirely. -kmp 9-Feb-2020
                msg_uuid = dictionary_lookup(msg_body, 'uuid')
                msg_sid = msg_body['sid']
                msg_curr_time = msg_body['timestamp']
                msg_detail = msg_body.get('detail')
                msg_telemetry = msg_body.get('telemetry_id')
                msg_diff = msg_body.get('diff', None)

                # build the object and index into ES
                # if strict, do not add uuids rev_linking to item to queue
                if msg_body['strict'] is True:
                    error = self.update_object(request, msg_uuid,
                                               add_to_secondary=None,
                                               sid=msg_sid, max_sid=max_sid,
                                               curr_time=msg_curr_time,
                                               telemetry_id=msg_telemetry)
                else:
                    error = self.update_object(request, msg_uuid,
                                               add_to_secondary=rev_linked_uuids,
                                               sid=msg_sid, max_sid=max_sid,
                                               curr_time=msg_curr_time,
                                               telemetry_id=msg_telemetry)
                if error:
                    if error.get('error_message') == 'defer_resend':
                        # resend the message and delete original so that receive
                        # count is not affected. set `deferred` to restart worker
                        to_defer.append(msg_body)
                        to_delete.append(msg)
                        deferred = True
                    elif error.get('error_message') == 'defer_replace':
                        # replace the message with a VisibilityTimeout
                        # set `deferred` to restart worker
                        self.queue.replace_messages([msg], target_queue=target_queue, vis_timeout=180)
                        deferred = True
                    else:
                        # regular error, replace the message with a VisibilityTimeout
                        # could do something with error, like putting on elasticache
                        self.queue.replace_messages([msg], target_queue=target_queue, vis_timeout=180)
                        errors.append(error)
                else:
                    # Sucessfully processed! (i.e. indexed or discarded conflict)
                    # if non-strict, adding will queue associated items to secondary
                    if msg_body['strict'] is False:
                        non_strict_uuids.add(msg_uuid)
                    counter[0] += 1  # do not increment on error
                    to_delete.append(msg)

                # delete messages when we have the right number
                if len(to_delete) == self.queue.delete_batch_size:
                    self.queue.delete_messages(to_delete, target_queue=target_queue)
                    to_delete = []

                # CHANGE - this needs to happen PER MESSAGE now
                # add to secondary queue, if applicable
                # search for all items that linkTo the non-strict items or contain
                # a rev_link to to them
                if non_strict_uuids or rev_linked_uuids:
                    queued, failed = self.find_and_queue_secondary_items(non_strict_uuids,  # THIS IS NOW A SINGLE UUID
                                                                         rev_linked_uuids,
                                                                         msg_sid,
                                                                         msg_telemetry,
                                                                         diff=msg_diff)
                    if failed:
                        error_msg = 'Failure(s) queueing secondary uuids: %s' % str(failed)
                        log.error('INDEXER: ', error=error_msg)
                        errors.append({'error_message': error_msg})
                    non_strict_uuids = set()
                    rev_linked_uuids = set()

            # if we need to restart the worker, break out of while loop
            if deferred:
                if to_defer:
                    self.queue.send_messages(to_defer, target_queue=target_queue)
                break

            # obtain more messages, possibly from a different queue
            prev_target_queue = target_queue
            messages, target_queue = self.get_messages_from_queue()
            # if we have switched between primary and secondary queues, delete
            # outstanding messages using previous queue
            if prev_target_queue != target_queue and to_delete:
                self.queue.delete_messages(to_delete, target_queue=prev_target_queue)
                to_delete = []

        # we're done. delete any outstanding messages before returning
        if to_delete:
            self.queue.delete_messages(to_delete, target_queue=target_queue)
        return errors, deferred

    def update_objects_sync(self, request, sync_uuids, counter):
        """
        Used with sync uuids (simply loop through)
        sync_uuids is a list of string uuids. Use timestamp of index run
        all uuids behave here as strict == true

        NOTE: This method does NOT take care of invalidation of items linked to
        indexed items. It is assumed that ALL items you want to index, including
        linked items are passed through the `sync_uuids` parameter
        """
        errors = []
        for i, uuid in enumerate(sync_uuids):
            error = self.update_object(request, uuid)
            if error is not None:
                errors.append(error)
            else:
                counter[0] += 1  # don't increment counter on an error
        return errors

    def update_object(self, request, uuid, add_to_secondary=None, sid=None,
                      max_sid=None, curr_time=None, telemetry_id=None):
        """
        Actually index the uuid using the index-data view.
        add_to_secondary is a set that gets the rev_linked_to_me
        from the request.embed(/<uuid>/@@index-data)
        """
        # logging constant
        cat = 'index object'

        # timing stuff
        start = timer()
        if not curr_time:
            curr_time = datetime.datetime.utcnow().isoformat()  # utc

        # to add to each log message
        log.bind(item_uuid=uuid, sid=sid, uo_start_time=curr_time)
        cm_source = False
        if telemetry_id:
            log.bind(telemetry_id=telemetry_id)
            # see if this message was generated by create-mapping
            if telemetry_id.startswith('cm_run_'):
                cm_source = True

        index_data_query = '/%s/@@index-data' % uuid
        try:
            # check sid first against max_sid from `update_objects_queue`
            # Raises SidException if invalid
            if sid and max_sid:
                check_sid(sid, max_sid)

            # invoke subrequest to get contents to index
            result = request.embed(index_data_query, as_user='INDEXER')
            duration = timer() - start
            # add total duration to indexing_stats in document
            result['indexing_stats']['total_indexing_view'] = duration
            log.bind(collection=result.get('item_type'))
            # log.info("Time for index-data", duration=duration, cat="indexing view")
        except SidException as e:
            duration = timer() - start
            log.warning('Invalid max sid. Resending...', duration=duration, cat=cat)
            # causes the item to be deferred by restarting worker
            # item will be re-sent (won't affect receive count)
            return {'error_message': 'defer_resend'}
        except MissingIndexItemException:
            # cannot find item. This could be due to it being purged.
            # if message is from create mapping, simply skip.
            # otherwise replace message and item will possibly make it to DLQ
            duration = timer() - start
            if cm_source:
                log.error('MissingIndexItemException encountered on resource %s'
                          ' from create_mapping. Skipping...' % index_data_query,
                          duration=duration, cat=cat)
                return
            else:
                log.warning('MissingIndexItemException encountered on resource '
                            '%s. No sid found. Replacing...' % index_data_query,
                            duration=duration, cat=cat)
                return {'error_message': 'defer_replace'}
        except Exception as e:
            duration = timer() - start
            log.error('Error rendering @@index-data', duration=duration, exc_info=True, cat=cat)
            return {'error_message': repr(e), 'time': curr_time, 'uuid': str(uuid)}

        # add found uuids that rev_link this item to be put in the secondary queue
        # find_and_queue_secondary_items() serves to find rev_linking items that
        # are currently in ES; this will pick up new rev links as well
        if add_to_secondary is not None:
            add_to_secondary.update(result['rev_linked_to_me'])

        last_exc = None
        for backoff in [0, 1, 2]:
            time.sleep(backoff)
            try:
                namespaced_index = get_namespaced_index(request, result['item_type'])
                self.es.index(
                    index=namespaced_index, doc_type=result['item_type'], body=result,
                    id=str(uuid), version=result['sid'], version_type='external_gte',
                    request_timeout=30
                )
            except ConflictError:
                # sid of found document is greater than sid of indexed document
                # this may be somewhat common and is not harmful
                # do not return an error so item is removed from queue
                duration = timer() - start
                log.warning('Conflict indexing', sid=result['sid'], duration=duration, cat=cat)
                return
            except (ConnectionError, ReadTimeoutError, TransportError) as e:
                duration = timer() - start
                log.warning('Retryable error indexing', error=str(e), duration=duration, cat=cat)
                last_exc = repr(e)
            except Exception as e:
                duration = timer() - start
                log.error('Error indexing', duration=duration, exc_info=True, cat=cat)
                last_exc = repr(e)
                break
            else:
                # success! Do not return an error so item is removed from queue
                duration = timer() - start
                log.info('Time to index', duration=duration, cat=cat)
                return
        # returning an error message means item did not index
        return {'error_message': last_exc, 'time': curr_time, 'uuid': str(uuid)}



from base64 import b64decode
import pytest
import webtest


# Test for blob storage

RED_DOT = """data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA
AAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO
9TXL0Y4OHwAAAABJRU5ErkJggg=="""

BLUE_DOT = """data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAA
oAAAAKAQMAAAC3/F3+AAAACXBIWXMAAA7DAAAOwwHHb6hkAA
AAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAgY0hSTQ
AAeiYAAICEAAD6AAAAgOgAAHUwAADqYAAAOpgAABdwnLpRPA
AAAANQTFRFALfvPEv6TAAAAAtJREFUCB1jYMAHAAAeAAEBGN
laAAAAAElFTkSuQmCC"""


@pytest.fixture
def testing_download(testapp):
    url = '/testing-downloads/'
    item = {
        'attachment': {
            'download': 'red-dot.png',
            'href': RED_DOT,
        },
        'attachment2': {
            'download': 'blue-dot.png',
            'href': BLUE_DOT,
        },
    }
    res = testapp.post_json(url, item, status=201)
    return res.location


def test_download_create(testapp, testing_download):
    res = testapp.get(testing_download)
    attachment = res.json['attachment']
    attachment2 = res.json['attachment2']

    assert attachment['href'] == '@@download/attachment/red-dot.png'
    assert attachment['type'] == 'image/png'
    assert attachment['width'] == 5
    assert attachment['height'] == 5
    assert attachment['md5sum'] == 'b60ab2708daec7685f3d412a5e05191a'
    url = testing_download + '/' + attachment['href']
    res = testapp.get(url)
    assert res.content_type == 'image/png'
    assert res.body == b64decode(RED_DOT.split(',', 1)[1])

    assert attachment2['href'] == '@@download/attachment2/blue-dot.png'
    assert attachment2['type'] == 'image/png'
    assert attachment2['width'] == 10
    assert attachment2['height'] == 10
    assert attachment2['md5sum'] == '013f03aa088adb19aa226c3439bda179'
    url = testing_download + '/' + attachment2['href']
    res = testapp.get(url)
    assert res.content_type == 'image/png'
    assert res.body == b64decode(BLUE_DOT.split(',', 1)[1])


def test_download_update(testapp, testing_download):
    item = {
        'attachment': {
            'download': 'blue-dot.png',
            'href': BLUE_DOT,
        },
        'attachment2': {
            'download': 'red-dot.png',
            'href': RED_DOT,
        },
    }
    testapp.put_json(testing_download, item, status=200)
    res = testapp.get(testing_download)
    attachment = res.json['attachment']
    attachment2 = res.json['attachment2']

    assert attachment['href'] == '@@download/attachment/blue-dot.png'
    url = testing_download + '/' + attachment['href']
    res = testapp.get(url)
    assert res.content_type == 'image/png'
    assert res.body == b64decode(BLUE_DOT.split(',', 1)[1])

    assert attachment2['href'] == '@@download/attachment2/red-dot.png'
    url = testing_download + '/' + attachment2['href']
    res = testapp.get(url)
    assert res.content_type == 'image/png'
    assert res.body == b64decode(RED_DOT.split(',', 1)[1])


def test_download_update_no_change(testapp, testing_download):
    item = {
        'attachment': {
            'download': 'red-dot.png',
            'href': '@@download/attachment/red-dot.png',
        },
        'attachment2': {
            'download': 'blue-dot.png',
            'href': '@@download/attachment2/blue-dot.png',
        },
    }
    testapp.put_json(testing_download, item, status=200)

    res = testapp.get(testing_download)
    attachment = res.json['attachment']
    attachment2 = res.json['attachment2']
    assert attachment['href'] == '@@download/attachment/red-dot.png'
    assert attachment2['href'] == '@@download/attachment2/blue-dot.png'


def test_download_update_one(testapp, testing_download):
    item = {
        'attachment': {
            'download': 'red-dot.png',
            'href': '@@download/attachment/red-dot.png',
        },
        'attachment2': {
            'download': 'red-dot.png',
            'href': RED_DOT,
        },
    }
    testapp.put_json(testing_download, item, status=200)

    res = testapp.get(testing_download)
    attachment = res.json['attachment']
    attachment2 = res.json['attachment2']

    assert attachment['href'] == '@@download/attachment/red-dot.png'
    url = testing_download + '/' + attachment['href']
    res = testapp.get(url)
    assert res.content_type == 'image/png'
    assert res.body == b64decode(RED_DOT.split(',', 1)[1])

    assert attachment2['href'] == '@@download/attachment2/red-dot.png'
    url = testing_download + '/' + attachment2['href']
    res = testapp.get(url)
    assert res.content_type == 'image/png'
    assert res.body == b64decode(RED_DOT.split(',', 1)[1])


def test_download_update_with_premade_href(testapp):
    """
    Test the functionality of PATCHing the attachment information directly
    (which is usually generated by the propsheet)
    """
    post_res = testapp.post_json('/testing-downloads/', {}, status=201)
    item_uuid = post_res.json['@graph'][0]['uuid']
    res = testapp.get('/testing-downloads/' + item_uuid).follow()
    # downloading will not work!
    with pytest.raises(webtest.AppError) as excinfo:
        testapp.get('/testing-downloads/' + item_uuid + '/@@download/attachment/blue-dot.png')
    assert 'Cannot find downloads' in str(excinfo.value)

    assert 'attachment' not in res.json
    patch_body = {
        'attachment': {
            'download': 'blue-dot.png',
            'href': '@@download/attachment/blue-dot.png',
        }
    }
    res2 = testapp.patch_json('/testing-downloads/' + item_uuid, patch_body, status=200)
    attachment = res2.json['@graph'][0]['attachment']
    assert attachment['href'] == '@@download/attachment/blue-dot.png'


def test_download_remove_one(testapp, testing_download):
    item = {
        'attachment': {
            'download': 'red-dot.png',
            'href': '@@download/attachment/red-dot.png',
        },
    }
    testapp.put_json(testing_download, item, status=200)

    res = testapp.get(testing_download)
    assert 'attachment' in res.json
    assert 'attachment2' not in res.json

    url = testing_download + '/@@download/attachment2/red-dot.png'
    res = testapp.get(url, status=404)


@pytest.mark.parametrize(
    'href',
    [
        '@@download/attachment/another.png',
        'http://example.com/another.png',
    ])
def test_download_update_bad_change(testapp, testing_download, href):
    item = {'attachment': {
        'download': 'red-dot.png',
        'href': href,
    }}
    testapp.put_json(testing_download, item, status=422)


@pytest.mark.parametrize(
    'href',
    [
        'http://example.com/another.png',
        'data:image/png;base64,NOT_BASE64',
        'data:image/png;NOT_A_PNG',
        'data:text/plain;asdf',
    ])
def test_download_create_bad_change(testapp, href):
    url = '/testing-downloads/'
    item = {'attachment': {
        'download': 'red-dot.png',
        'href': href,
    }}
    testapp.post_json(url, item, status=422)


def test_download_create_wrong_extension(testapp):
    url = '/testing-downloads/'
    item = {'attachment': {
        'download': 'red-dot.jpg',
        'href': RED_DOT,
    }}
    testapp.post_json(url, item, status=422)


def test_download_create_w_wrong_md5sum(testapp):
    url = '/testing-downloads/'
    item = {'attachment': {
        'download': 'red-dot.jpg',
        'href': RED_DOT,
        'md5sum': 'deadbeef',
    }}
    testapp.post_json(url, item, status=422)
